
#include "Widget.h"
#include "WidgetDriver.h"
#include "WidgetMessage.h"
#include "WidgetRelatedObject.h"
#include "WidgetLayoutChildren.h"
#include <cassert>
#include <algorithm>
#include <functional>

namespace ghost{

	namespace{

		ObjectPool<Widget>& GetWidgetPool_()
		{
			static ObjectPool<Widget> pool;
			return pool;
		}

	} // namespace

	typedef std::set<Widget*> WidgetSet;

	class WidgetImpl_{
		friend class Widget;

		widget::RelatedObject* pRelatedObject_;

		widget::Driver_* pDriver_;

		Widget* pParent_;
		WidgetSet children_;

		RECT absoluteRect_; // ڿͻľ

#ifdef _DEBUG
		COLORREF testFrameColor_; // չֿؼı߿ɫ
#endif // _DEBUG

	private:
		WidgetImpl_::WidgetImpl_()
			: pRelatedObject_(0)
			, pDriver_(0)
			, pParent_(0)
#ifdef _DEBUG
			, testFrameColor_(RGB(rand()%256, rand()%256, rand()%256))
#endif // _DEBUG
		{
			::SetRectEmpty(&absoluteRect_);
		}
	};

	Widget::Widget(HWND hWnd) 
		: pImpl_(new WidgetImpl_)
	{
		pImpl_->pRelatedObject_ = new widget::RelatedObject(this);
		pImpl_->pDriver_ = new widget::Driver_(hWnd);
		pImpl_->pDriver_->SetRootWidget(this);

		// Ϣ
		SendMessage(widget::MSG_CREATE);
	}

	Widget::Widget(Widget* pParent) 
		: pImpl_(new WidgetImpl_)
	{
		SetParent(pParent);
	}

	Widget::~Widget()
	{
		// Ϣ
		SendMessage(widget::MSG_DESTROY);

		// ӿؼ
		WidgetSet temp(std::move(pImpl_->children_));
		std::for_each(temp.begin(), temp.end(), std::mem_fn(&Widget::Destroy));

		if (IsRoot()) 
		{
			// Ϊؼͬʱ
			delete pImpl_->pDriver_;
		}
		else
		{
			// Ǹؼ븸ؼ
			SetParent(0);
		}

		delete pImpl_->pRelatedObject_;
		delete pImpl_;
	}

	Widget* Widget::Create(HWND hWnd)
	{
		return GetWidgetPool_().Construct(hWnd);
	}

	Widget* Widget::Create(Widget* pParent)
	{
		assert(pParent);
		if (pParent)
		{
			return GetWidgetPool_().Construct(pParent);
		}
		return 0;
	}

	void Widget::Destroy()
	{
		GetWidgetPool_().Destroy(this);
	}

	void Widget::RelationshipCreated_(ObjectRelationship* pObject)
	{
		GetRelatedObject()->RelationshipCreated_(pObject);
	}
	void Widget::RelationshipDestroyed_(ObjectRelationship* pObject)
	{
		GetRelatedObject()->RelationshipDestroyed_(pObject);
	}

	widget::RelatedObject* Widget::GetRelatedObject() const
	{
		return pImpl_->pRelatedObject_;
	}

	HWND Widget::GetContainerWindow() const
	{
		return pImpl_->pDriver_->GetContainerWindow();
	}

	Widget* Widget::GetRoot() const
	{
		return pImpl_->pDriver_->GetRootWidget();
	}

	bool Widget::IsRoot() const
	{
		return pImpl_->pDriver_->GetRootWidget() == this;
	}

	void Widget::SetParent(Widget* const pNewParent)
	{
		// ǷѾǸӹϵ
		if (pNewParent == pImpl_->pParent_)
		{
			return;
		}

		Widget* pOldParent = pImpl_->pParent_;
		pImpl_->pParent_ = 0; // ؼRemoveChildԴ˶SetParent(0)Ϊ0
		if (pOldParent)
		{
			pOldParent->RemoveChild(this);
		}
		pImpl_->pParent_ = pNewParent; // ؼInsertChildԴ˶SetParent(ָ)Ϊؼָ
		if (pNewParent) 
		{
			// ԼԼӿؼΪ¸ؼ
			auto pNewDriver = pNewParent->pImpl_->pDriver_;
			pImpl_->pDriver_ = pNewDriver;
			std::for_each(pImpl_->children_.begin(), pImpl_->children_.end(), [pNewDriver](WidgetSet::value_type pChild){
				pChild->pImpl_->pDriver_ = pNewDriver;
			});
			pNewParent->InsertChild(this);
		}
		else
		{
			// ԼԼӿؼΪ0
			pImpl_->pDriver_ = 0;
			std::for_each(pImpl_->children_.begin(), pImpl_->children_.end(), [](WidgetSet::value_type pChild){
				pChild->pImpl_->pDriver_ = 0;
			});
		}
	}

	Widget* Widget::GetParent() const
	{
		return pImpl_->pParent_;
	}

	bool Widget::InsertChild(Widget* const pChild)
	{
		assert(pChild);
		if (!pChild)
		{
			return false;
		}
		
		bool ret = pImpl_->children_.insert(pChild).second;
		pChild->SetParent(this);
		return ret;
	}
	bool Widget::RemoveChild(Widget* const pChild)
	{
		if (0 < pImpl_->children_.erase(pChild))
		{
			assert(pChild);
			pChild->SetParent(0);
			return true;
		}
		return false;
	}

	void Widget::SetAbsoluteRect(const RECT& rect, bool update/* = true*/)
	{
		if (::EqualRect(&pImpl_->absoluteRect_, &rect))
		{
			return;
		}
		if (update)
		{
			HRGN hRgn = ::CreateRectRgnIndirect(&pImpl_->absoluteRect_);
			HRGN hRgn1 = ::CreateRectRgnIndirect(&rect);
			::CombineRgn(hRgn, hRgn, hRgn1, RGN_OR);

			::CopyRect(&pImpl_->absoluteRect_, &rect);
			// 
			if (pImpl_->pDriver_)
			{
				pImpl_->pDriver_->UpdateWindow(hRgn);
			}

			::DeleteObject(hRgn1);
			::DeleteObject(hRgn);
		}
		else
		{
			::CopyRect(&pImpl_->absoluteRect_, &rect);
		}
		// ӿؼв
		LayoutChildren();
		// ıϢ
		SendMessage(widget::MSG_RECT_CHANGED);
	}
	void Widget::GetAbsoluteRect(RECT& rect)
	{
		::CopyRect(&rect, &pImpl_->absoluteRect_);
	}

	void Widget::Draw(HDC hDC)
	{
		SendMessage(widget::MSG_DRAW, hDC);
		// ӿؼ
		DrawChildren(hDC);
	}

	void Widget::LayoutChildren()
	{
		auto pRelatedObject = GetRelatedObject();
		if (pRelatedObject)
		{
			const widget::LayoutChildrenSet& layoutChildrens = pRelatedObject->GetLayoutChildrens();
			std::for_each(
				layoutChildrens.begin(), layoutChildrens.end(), 
				std::bind(std::mem_fn(&widget::LayoutChildren::Layout), std::placeholders::_1, this));
		}
	}

	void Widget::DrawChildren(HDC hDC)
	{
		std::for_each(
			pImpl_->children_.begin(), pImpl_->children_.end(), 
			std::bind(std::mem_fn(&Widget::Draw), std::placeholders::_1, hDC));
	}

	int Widget::SendMessage(widget::Message message, void* param1/* = 0*/, void* param2/* = 0*/)
	{
		switch (message)
		{
#ifdef _DEBUG
		case widget::MSG_DRAW:
			{
				HBRUSH hBrush = ::CreateSolidBrush(pImpl_->testFrameColor_);
				::FrameRect((HDC)param1, &pImpl_->absoluteRect_, hBrush);
				::DeleteObject(hBrush);
			}
			break;
#endif // _DEBUG
		}
		return widget::MSG_RESULT_NONE;
	}

} // namespace ghost